"use strict";
/*
 * decaffeinate suggestions:
 * DS101: Remove unnecessary use of Array.from
 * DS102: Remove unnecessary code created because of implicit returns
 * DS205: Consider reworking code to avoid use of IIFEs
 * DS207: Consider shorter variations of null checks
 * Full docs: https://github.com/decaffeinate/decaffeinate/blob/master/docs/suggestions.md
 */
Object.defineProperty(exports, "__esModule", { value: true });
var scoring_1 = require("./scoring");
var types_1 = require("./types");
var feedback = {
    default_feedback: {
        warning: null,
        suggestions: [
            'Use a few words, avoid common phrases',
            'No need for symbols, digits, or uppercase letters',
        ],
    },
    get_feedback: function (score, sequence) {
        // starting feedback
        if (sequence.length === 0) {
            return feedback.default_feedback;
        }
        // no feedback if score is good or great.
        if (score > 2) {
            return {
                warning: null,
                suggestions: [],
            };
        }
        // tie feedback to the longest match for longer sequences
        var longest_match = sequence[0];
        for (var _i = 0, _a = sequence.slice(1); _i < _a.length; _i++) {
            var match = _a[_i];
            if (match.token.length > longest_match.token.length) {
                longest_match = match;
            }
        }
        var f = feedback.get_match_feedback(longest_match, sequence.length === 1);
        var extra_feedback = 'Add another word or two. Uncommon words are better.';
        if (f != null) {
            f.suggestions.unshift(extra_feedback);
            if (f.warning == null) {
                f.warning = null;
            }
        }
        else {
            f = {
                warning: null,
                suggestions: [extra_feedback],
            };
        }
        return f;
    },
    get_match_feedback: function (match, is_sole_match) {
        switch (match.pattern) {
            case 'dictionary':
                return feedback.get_dictionary_match_feedback(match, is_sole_match);
            case 'spatial':
                var warning = match.turns === 1
                    ? types_1.Warning.SpatialStraightRow
                    : types_1.Warning.SpatialShortKeyboardPattern;
                return {
                    warning: warning,
                    suggestions: ['Use a longer keyboard pattern with more turns'],
                };
            case 'repeat':
                warning =
                    match.base_token.length === 1
                        ? types_1.Warning.RepeatCharacter
                        : types_1.Warning.RepeatPattern;
                return {
                    warning: warning,
                    suggestions: ['Avoid repeated words and characters'],
                };
            case 'sequence':
                return {
                    warning: types_1.Warning.Sequence,
                    suggestions: ['Avoid sequences'],
                };
            case 'regex':
                if (match.regex_name === 'recent_year') {
                    return {
                        warning: types_1.Warning.RegexYear,
                        suggestions: [
                            'Avoid recent years',
                            'Avoid years that are associated with you',
                        ],
                    };
                }
                break;
            case 'date':
                return {
                    warning: types_1.Warning.Date,
                    suggestions: ['Avoid dates and years that are associated with you'],
                };
        }
    },
    get_dictionary_match_feedback: function (match, is_sole_match) {
        var warning = (function () {
            if (match.dictionary_name === 'passwords') {
                if (is_sole_match && !match.l33t && !match.reversed) {
                    return types_1.Warning.CommonPassword;
                }
                else if (match.guesses_log10 <= 4) {
                    return types_1.Warning.SimilarToCommonPassword;
                }
            }
            else {
                return null;
            }
        })();
        var suggestions = [];
        var word = match.token;
        if (word.match(scoring_1.default.START_UPPER)) {
            suggestions.push("Capitalization doesn't help very much");
        }
        else if (word.match(scoring_1.default.ALL_UPPER) && word.toLowerCase() !== word) {
            suggestions.push('All-uppercase is almost as easy to guess as all-lowercase');
        }
        if (match.reversed && match.token.length >= 4) {
            suggestions.push("Reversed words aren't much harder to guess");
        }
        if (match.l33t) {
            suggestions.push("Predictable substitutions like '@' instead of 'a' don't help very much");
        }
        var result = {
            warning: warning,
            suggestions: suggestions,
        };
        return result;
    },
};
exports.default = feedback;
//# sourceMappingURL=feedback.js.map